/*
 * Name:    d_network_relay.c
 *
 * Purpose: MX driver for network relay support.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_record.h"
#include "mx_driver.h"
#include "mx_net.h"
#include "mx_relay.h"
#include "d_network_relay.h"

/* Initialize the network relay driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_network_relay_record_function_list = {
	NULL,
	mxd_network_relay_create_record_structures,
};

MX_RELAY_FUNCTION_LIST mxd_network_relay_relay_function_list = {
	mxd_network_relay_relay_command,
	mxd_network_relay_get_relay_status
};

MX_RECORD_FIELD_DEFAULTS mxd_network_relay_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_RELAY_STANDARD_FIELDS,
	MXD_NETWORK_RELAY_STANDARD_FIELDS
};

long mxd_network_relay_num_record_fields
	= sizeof( mxd_network_relay_record_field_defaults )
		/ sizeof( mxd_network_relay_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_relay_rfield_def_ptr
			= &mxd_network_relay_record_field_defaults[0];

/*=======================================================================*/

static mx_status_type
mxd_network_relay_get_pointers( MX_RELAY *relay,
			MX_NETWORK_RELAY **network_relay,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_relay_get_pointers()";

	if ( relay == (MX_RELAY *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The relay pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( relay->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_RECORD pointer for relay pointer passed by '%s' is NULL.",
			calling_fname );
	}

	if ( relay->record->mx_type != MXT_RLY_NETWORK ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The relay '%s' passed by '%s' is not a network relay.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			relay->record->name, calling_fname,
			relay->record->mx_superclass,
			relay->record->mx_class,
			relay->record->mx_type );
	}

	if ( network_relay != (MX_NETWORK_RELAY **) NULL ) {

		*network_relay = (MX_NETWORK_RELAY *)
				(relay->record->record_type_struct);

		if ( *network_relay == (MX_NETWORK_RELAY *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_NETWORK_RELAY pointer for relay record '%s' passed by '%s' is NULL",
				relay->record->name, calling_fname );
		}
	}

	if ( (*network_relay)->server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX server record pointer for MX network relay '%s' is NULL.",
			relay->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mxd_network_relay_create_record_structures( MX_RECORD *record )
{
        const char fname[] = "mxd_network_relay_create_record_structures()";

        MX_RELAY *relay;
        MX_NETWORK_RELAY *network_relay;

        /* Allocate memory for the necessary structures. */

        relay = (MX_RELAY *) malloc(sizeof(MX_RELAY));

        if ( relay == (MX_RELAY *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_RELAY structure." );
        }

        network_relay = (MX_NETWORK_RELAY *)
				malloc( sizeof(MX_NETWORK_RELAY) );

        if ( network_relay == (MX_NETWORK_RELAY *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_NETWORK_RELAY structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = relay;
        record->record_type_struct = network_relay;
        record->class_specific_function_list
			= &mxd_network_relay_relay_function_list;

        relay->record = record;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_relay_relay_command( MX_RELAY *relay )
{
	const char fname[] = "mxd_network_relay_relay_command()";

	MX_NETWORK_RELAY *network_relay;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	mx_status_type mx_status;

	mx_status = mxd_network_relay_get_pointers(
					relay, &network_relay, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf(rfname, "%s.relay_command", network_relay->remote_record_name);

	mx_status = mx_put( network_relay->server_record,
				rfname, MXFT_INT, &(relay->relay_command) );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_relay_get_relay_status( MX_RELAY *relay )
{
	const char fname[] = "mxd_network_relay_get_relay_status()";

	MX_NETWORK_RELAY *network_relay;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int relay_status;
	mx_status_type mx_status;

	mx_status = mxd_network_relay_get_pointers(
					relay, &network_relay, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf(rfname, "%s.relay_status", network_relay->remote_record_name);

	mx_status = mx_get( network_relay->server_record,
				rfname, MXFT_INT, &relay_status );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( relay_status == MXF_OPEN_RELAY ) {
		relay->relay_status = MXF_OPEN_RELAY;
	} else {
		relay->relay_status = MXF_CLOSE_RELAY;
	}

	return MX_SUCCESSFUL_RESULT;
}

