/*
 * Name:    mx_socket.h 
 *
 * Purpose: BSD sockets and Winsock interface functions.
 *
 * Author:  William Lavender
 *
 * This header file includes all the defines and prototypes normally needed
 * by a socket-based program.  Thus, to use sockets, it is only necessary to
 * use #include "mx_socket.h" rather than including all the individual headers
 * separately.  This also promotes portability since different operating
 * systems have different sets of header files that they want to be included
 * in order to implement socket functionality.
 *
 *----------------------------------------------------------------------
 *
 * Copyright 1999-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef __MX_SOCKET_H__
#define __MX_SOCKET_H__

#include "mx_osdef.h"

/* Include standard socket defines and prototypes. */

#if defined( OS_WIN32 )
#  include <winsock.h>

#elif defined( OS_VMS )

#  include "vms_twg.h"		/* For WIN/TCP from The Wollogong Group. */

#else

#  include <netdb.h>
#  include <sys/time.h>
#  include <sys/types.h>
#  include <sys/socket.h>
#  include <sys/ioctl.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>

#  if defined( OS_HPUX )
      extern int h_errno;
#  endif

#  if defined( OS_DJGPP )
      extern int sock_init(void);
      extern int h_errno;
#  endif

#endif

/* ===== Unix BSD sockets support ===== */

#if defined( OS_UNIX ) || defined( OS_CYGWIN ) || defined( OS_VMS ) \
 || defined( OS_DJGPP )

#  define MX_SOCKET			int
#  define MX_INVALID_SOCKET		(-1)
#  define MX_SOCKET_ERROR		(-1)

#  define mx_socket_get_last_error()	(errno)

/* ===== Microsoft Win32 Winsock support ===== */

#elif defined( OS_WIN32 )

#  define MX_SOCKET			SOCKET
#  define MX_INVALID_SOCKET		INVALID_SOCKET
#  define MX_SOCKET_ERROR		SOCKET_ERROR

#  define mx_socket_get_last_error()	WSAGetLastError()

/* ============================ */

#else
#  error "Socket definitions not configured for this operating system."
#endif

/* ***** Various define's ***** */

/* The following defines are for the second argument to the function
 * mx_socket_check_error_status().
 */

#define MXF_SOCKCHK_INVALID	(-1)
#define MXF_SOCKCHK_ZERO	0
#define MXF_SOCKCHK_NULL	1

/* The following are flags for the mx_socket_send() and mx_socket_receive()
 * functions.
 */

#define MXF_SOCKET_USE_LINE_TERMINATORS		0x1

/* ***** General function definitions ***** */

MX_API mx_status_type mx_socket_open_as_client( MX_SOCKET *client_socket,
						char *hostname,
						int port_number,
						int open_flags );

MX_API mx_status_type mx_socket_open_as_server( MX_SOCKET *server_socket,
						int port_number,
						int open_flags );

MX_API mx_status_type mx_socket_close( MX_SOCKET tcp_socket );

MX_API int  mx_socket_is_open( MX_SOCKET tcp_socket );

MX_API void mx_socket_mark_as_closed( MX_SOCKET *tcp_socket );

MX_API char *mx_socket_strerror( int saved_errno );

MX_API int  mx_socket_check_error_status( void *value_to_check,
						int type_of_check,
						char **error_string );

MX_API mx_status_type mx_socket_get_inet_address( char *hostname,
						unsigned long *inet_address );

MX_API mx_status_type mx_socket_send( MX_SOCKET mx_socket,
					void *message_buffer,
					size_t message_length_in_bytes,
					int flags );

MX_API mx_status_type mx_socket_receive( MX_SOCKET mx_socket,
					void *message_buffer,
					size_t message_length_in_bytes,
					int flags );

MX_API mx_status_type mx_socket_putline( MX_SOCKET mx_socket, char *buffer );

MX_API mx_status_type mx_socket_getline( MX_SOCKET mx_socket,
						char *buffer,
						size_t buffer_length );

#endif /* __MX_SOCKET_H__ */

